from pathlib import Path
from typing import Dict, Optional, Union

import pytest
from pydantic_settings import SettingsConfigDict

from ape.api import ConfigEnum, PluginConfig
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.managers.config import (
    CONFIG_FILE_NAME,
    ConfigManager,
    DeploymentConfigCollection,
    merge_configs,
)
from ape.types import GasLimit
from ape.utils import create_tempdir
from ape_ethereum.ecosystem import NetworkConfig
from ape_networks import CustomNetwork
from tests.functional.conftest import PROJECT_WITH_LONG_CONTRACTS_FOLDER


def test_deployments(networks_connected_to_tester, owner, vyper_contract_container, config):
    networks = networks_connected_to_tester  # Connection needs to lookup config.

    # First, obtain a "previously-deployed" contract.
    instance = vyper_contract_container.deploy(1000200000, sender=owner)
    address = instance.address

    # Create a config using this new contract for a "later time".
    data = {
        **_create_deployments(address=address, contract_name=instance.contract_type.name),
        "valid_ecosystems": {"ethereum": networks.ethereum},
        "valid_networks": [LOCAL_NETWORK_NAME],
    }
    deploy_config = DeploymentConfigCollection(root=data)
    assert deploy_config.root["ethereum"]["local"][0]["address"] == address

    orig = config.deployments
    config.deployments = deploy_config
    try:
        # Ensure we can reference the deployment on the contract.
        deployment = vyper_contract_container.deployments[0]
    finally:
        config.deployments = orig

    assert deployment.address == instance.address


def test_deployments_integer_type_addresses(networks):
    data = {
        **_create_deployments(address=0x0C25212C557D00024B7CA3DF3238683A35541354),
        "valid_ecosystems": {"ethereum": networks.ethereum},
        "valid_networks": [LOCAL_NETWORK_NAME],
    }
    config = DeploymentConfigCollection(root=data)
    assert (
        config.root["ethereum"]["local"][0]["address"]
        == "0x0c25212c557d00024b7Ca3df3238683A35541354"
    )


@pytest.mark.parametrize(
    "ecosystem_names,network_names,err_part",
    [(["ERRORS"], ["mainnet"], "ecosystem"), (["ethereum"], ["ERRORS"], "network")],
)
def test_deployments_bad_value(
    ecosystem_names, network_names, err_part, ape_caplog, plugin_manager
):
    deployments = _create_deployments()
    all_ecosystems = dict(plugin_manager.ecosystems)
    ecosystem_dict = {e: all_ecosystems[e] for e in ecosystem_names if e in all_ecosystems}
    data = {**deployments, "valid_ecosystems": ecosystem_dict, "valid_networks": network_names}
    ape_caplog.assert_last_log_with_retries(
        lambda: DeploymentConfigCollection(root=data),
        f"Invalid {err_part}",
    )


def _create_deployments(
    ecosystem_name: str = "ethereum",
    network_name: str = "local",
    address: Union[int, str] = "0x0C25212C557D00024B7CA3DF3238683A35541354",
    contract_name: Optional[str] = "MyContract",
) -> Dict:
    return {
        ecosystem_name: {
            network_name: [
                {
                    "address": address,
                    "contract_type": contract_name,
                }
            ]
        }
    }


def test_ethereum_network_configs(config, temp_config):
    eth_config = {"ethereum": {"goerli": {"default_provider": "test"}}}
    with temp_config(eth_config):
        actual = config.get_config("ethereum")
        assert actual.goerli.default_provider == "test"

        # Ensure that non-updated fields remain unaffected
        assert actual.goerli.block_time == 15


def test_network_gas_limit_default(config):
    eth_config = config.get_config("ethereum")

    assert eth_config.goerli.gas_limit == "auto"
    assert eth_config.local.gas_limit == "max"


def _goerli_with_gas_limit(gas_limit: GasLimit) -> dict:
    return {
        "ethereum": {
            "goerli": {
                "default_provider": "test",
                "gas_limit": gas_limit,
            }
        }
    }


@pytest.mark.parametrize("gas_limit", ("auto", "max"))
def test_network_gas_limit_string_config(gas_limit, config, temp_config):
    eth_config = _goerli_with_gas_limit(gas_limit)

    with temp_config(eth_config):
        actual = config.get_config("ethereum")

        assert actual.goerli.gas_limit == gas_limit

        # Local configuration is unaffected
        assert actual.local.gas_limit == "max"


@pytest.mark.parametrize("gas_limit", (1234, "1234", 0x4D2, "0x4D2"))
def test_network_gas_limit_numeric_config(gas_limit, config, temp_config):
    eth_config = _goerli_with_gas_limit(gas_limit)

    with temp_config(eth_config):
        actual = config.get_config("ethereum")

        assert actual.goerli.gas_limit == 1234

        # Local configuration is unaffected
        assert actual.local.gas_limit == "max"


def test_network_gas_limit_invalid_numeric_string(config, temp_config):
    """
    Test that using hex strings for a network's gas_limit config must be
    prefixed with '0x'
    """
    eth_config = _goerli_with_gas_limit("4D2")
    with pytest.raises(ValueError, match="Gas limit hex str must include '0x' prefix."):
        with temp_config(eth_config):
            pass


def test_dependencies(project_with_dependency_config, config):
    assert len(config.dependencies) == 1
    assert config.dependencies[0].name == "testdependency"
    assert config.dependencies[0].contracts_folder == "source/v0.1"
    assert config.dependencies[0].local == str(PROJECT_WITH_LONG_CONTRACTS_FOLDER)


def test_config_access():
    config = NetworkConfig()
    assert "default_provider" in config
    assert (
        config.default_provider
        == config["default_provider"]
        == getattr(config, "default-provider")
        == "geth"
    )


def test_plugin_config_updates_when_default_is_empty_dict():
    class SubConfig(PluginConfig):
        foo: int = 0
        bar: int = 1

    class MyConfig(PluginConfig):
        sub: Dict[str, Dict[str, SubConfig]] = {}

    overrides = {"sub": {"baz": {"test": {"foo": 5}}}}
    actual = MyConfig.from_overrides(overrides)
    assert actual.sub == {"baz": {"test": SubConfig(foo=5, bar=1)}}


@pytest.mark.parametrize("override_0,override_1", [(True, {"foo": 0}), ({"foo": 0}, True)])
def test_plugin_config_with_union_dicts(override_0, override_1):
    class SubConfig(PluginConfig):
        bool_or_dict: Union[bool, Dict] = True
        dict_or_bool: Union[Dict, bool] = {}

    config = SubConfig.from_overrides({"bool_or_dict": override_0, "dict_or_bool": override_1})
    assert config.bool_or_dict == override_0
    assert config.dict_or_bool == override_1


def test_global_config(data_folder, config):
    config_file = data_folder / CONFIG_FILE_NAME
    config_file.unlink(missing_ok=True)
    config_file.touch()
    config_content = """
test:
  number_of_accounts: 11
""".strip()
    config_file.write_text(config_content)
    config.load(force_reload=True)
    assert config.get_config("test").number_of_accounts == 11
    config_file.unlink(missing_ok=True)


def test_merge_configs():
    """
    The test covers most cases in `merge_config()`. See comment below explaining
    `expected`.
    """
    global_config = {
        "ethereum": {
            "mainnet": {"default_provider": "geth"},
            "local": {"default_provider": "test", "required_confirmations": 5},
        }
    }
    project_config = {
        "ethereum": {
            "local": {"default_provider": "geth"},
            "sepolia": {"default_provider": "alchemy"},
        },
        "test": "foo",
    }
    actual = merge_configs(global_config, project_config)

    # Expected case `key only in global`: "mainnet"
    # Expected case `non-primitive override from project`: local -> default_provider, which
    #  is `test` in global but `geth` in project; thus it is `geth` in expected.
    # Expected case `merge sub-dictionaries`: `ethereum` is being merged.
    # Expected case `add missing project keys`: `test` is added, so is `sepolia` (nested-example).
    expected = {
        "ethereum": {
            "local": {"default_provider": "geth", "required_confirmations": 5},
            "mainnet": {"default_provider": "geth"},
            "sepolia": {"default_provider": "alchemy"},
        },
        "test": "foo",
    }
    assert actual == expected


def test_merge_configs_short_circuits():
    """
    Cover all short-circuit cases.
    """
    ex = {"test": "foo"}
    assert merge_configs({}, {}) == {}
    assert merge_configs(ex, {}) == merge_configs({}, ex) == ex


def test_plugin_config_getattr_and_getitem(config):
    config = config.get_config("ethereum")
    assert config.mainnet is not None
    assert config.mainnet == config["mainnet"]


def test_custom_plugin_config_extras():
    class CustomConfig(PluginConfig):
        model_config = SettingsConfigDict(extra="allow")

    config = CustomConfig(foo="123")
    assert "foo" in config
    assert config.foo == "123"
    assert config["foo"] == "123"


def test_config_enum():
    class MyEnum(ConfigEnum):
        FOO = "FOO"
        BAR = "BAR"

    class MyConfig(PluginConfig):
        my_enum: MyEnum

    actual = MyConfig(my_enum="FOO")
    assert actual.my_enum == MyEnum.FOO


def test_config_manager_loads_on_init(config):
    """
    This is needed or else tools may interact with the config manager
    before it has processed the config file.
    """
    name = "nametestvalidate"

    with create_tempdir() as path:
        config = f"name: {name}"
        (path / "ape-config.yaml").write_text(config)
        manager = ConfigManager(REQUEST_HEADER={}, DATA_FOLDER=Path.cwd(), PROJECT_FOLDER=path)
        assert manager.name == name


def test_load_does_not_call_project_manager(temp_config, config):
    """
    It is highly critical that `load()` does not call anything from
    `project_manager` as it is not loaded yet in the manager access mixin.
    """
    orig = config.project_manager.path
    path = Path("_should_not_be_in_parents_")
    try:
        with temp_config({"contracts_folder": "src"}):
            config.project_manager.path = path
            assert config.load(force_reload=True)
            assert path.name not in [x.name for x in config.contracts_folder.parents]
    finally:
        config.project_manager.path = orig


def test_contracts_folder_with_hyphen(temp_config):
    with temp_config({"contracts-folder": "src"}) as project:
        assert project.contracts_folder.name == "src"


def test_compiler_cache_folder(temp_config):
    with temp_config(
        {"contracts_folder": "smarts", "compile": {"cache_folder": ".cash"}}
    ) as project:
        assert project.contracts_folder.name == "smarts"
        assert project.compiler_cache_folder.name == ".cash"
        assert str(project.contracts_folder) not in str(project.compiler_cache_folder)


def test_custom_network():
    chain_id = 11191919191991918223773
    data = {
        "name": "mytestnet",
        "chain_id": 11191919191991918223773,
        "ecosystem": "ethereum",
    }
    network = CustomNetwork.model_validate(data)
    assert network.name == "mytestnet"
    assert network.chain_id == chain_id
    assert network.ecosystem == "ethereum"



from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, TypeVar

from pydantic import ConfigDict
from pydantic_settings import BaseSettings

from ape.utils.basemodel import _assert_not_ipython_check, only_raise_attribute_error

if TYPE_CHECKING:
    from ape.managers.config import ConfigManager

ConfigItemType = TypeVar("ConfigItemType")


class ConfigEnum(str, Enum):
    """
    A configuration `Enum <https://docs.python.org/3/library/enum.html>`__ type.
    Use this to limit the values of a config item, such as colors ``"RED"``, ``"BLUE"``,
    ``"GREEN"``, rather than any arbitrary ``str``.

    Usage example::

            class MyEnum(ConfigEnum):
                FOO = "FOO"
                BAR = "BAR"

            class MyConfig(PluginConfig):
                my_enum: MyEnum

            model = MyConfig(my_enum="FOO")

    """


class PluginConfig(BaseSettings):
    """
    A base plugin configuration class. Each plugin that includes
    a config API must register a subclass of this class.
    """

    # NOTE: This is probably partially initialized at the time of assignment
    _config_manager: Optional["ConfigManager"]

    @classmethod
    def from_overrides(
        cls, overrides: Dict, config_manager: Optional["ConfigManager"] = None
    ) -> "PluginConfig":
        cls._config_manager = config_manager
        default_values = cls().model_dump()

        def update(root: Dict, value_map: Dict):
            for key, val in value_map.items():
                if isinstance(val, dict) and key in root and isinstance(root[key], dict):
                    root[key] = update(root[key], val)
                else:
                    root[key] = val

            return root

        return cls(**update(default_values, overrides))

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        _assert_not_ipython_check(attr_name)

        # Allow hyphens in plugin config files.
        attr_name = attr_name.replace("-", "_")
        extra = self.__pydantic_extra__ or {}
        if attr_name in extra:
            return extra[attr_name]

        return super().__getattribute__(attr_name)

    def __getitem__(self, item: str) -> Any:
        extra = self.__pydantic_extra__ or {}
        if item in self.__dict__:
            return self.__dict__[item]

        elif item in extra:
            return extra[item]

        raise KeyError(f"'{item}' not in config.")

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__ or key in (self.__pydantic_extra__ or {})

    def get(self, key: str, default: Optional[ConfigItemType] = None) -> ConfigItemType:
        extra: Dict = self.__pydantic_extra__ or {}
        return self.__dict__.get(key, extra.get(key, default))


class GenericConfig(ConfigDict):
    """
    The default class used when no specialized class is used.
    """



import importlib
from typing import Any, Generator, Iterator, List, Optional, Set, Tuple

from ape.__modules__ import __modules__
from ape.exceptions import ApeAttributeError
from ape.logging import logger
from ape.plugins._utils import _filter_plugins_from_dists, clean_plugin_name
from ape.plugins.pluggy_patch import plugin_manager as pluggy_manager
from ape.utils.basemodel import _assert_not_ipython_check, only_raise_attribute_error
from ape.utils.misc import _get_distributions, log_instead_of_fail


def valid_impl(api_class: Any) -> bool:
    """
    Check if an API class is valid. The class must not have any unimplemented
    abstract methods.

    Args:
        api_class (any)

    Returns:
        bool
    """

    if isinstance(api_class, tuple):
        return all(valid_impl(c) for c in api_class)

    # Is not an ABC base class or abstractdataclass
    if not hasattr(api_class, "__abstractmethods__"):
        return True  # not an abstract class

    return len(api_class.__abstractmethods__) == 0


class PluginManager:
    _unimplemented_plugins: List[str] = []

    def __init__(self) -> None:
        self.__registered = False

    @log_instead_of_fail(default="<PluginManager>")
    def __repr__(self) -> str:
        return f"<{PluginManager.__name__}>"

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Iterator[Tuple[str, Tuple]]:
        _assert_not_ipython_check(attr_name)

        # NOTE: The first time this method is called, the actual
        #  plugin registration occurs. Registration only happens once.
        self._register_plugins()

        if not hasattr(pluggy_manager.hook, attr_name):
            raise ApeAttributeError(f"{PluginManager.__name__} has no attribute '{attr_name}'.")

        # Do this to get access to the package name
        hook_fn = getattr(pluggy_manager.hook, attr_name)
        hookimpls = hook_fn.get_hookimpls()

        def get_plugin_name_and_hookfn(h):
            return h.plugin_name, getattr(h.plugin, attr_name)()

        for plugin_name, results in map(get_plugin_name_and_hookfn, hookimpls):
            # NOTE: Some plugins return a tuple and some return iterators
            if not isinstance(results, Generator):
                validated_plugin = self._validate_plugin(plugin_name, results)
                if validated_plugin:
                    yield validated_plugin
            else:
                # Only if it's an iterator, provider results as a series
                for result in results:
                    validated_plugin = self._validate_plugin(plugin_name, result)
                    if validated_plugin:
                        yield validated_plugin

    @property
    def registered_plugins(self) -> Set[str]:
        self._register_plugins()
        return {x[0] for x in pluggy_manager.list_name_plugin()}

    def _register_plugins(self):
        if self.__registered:
            return

        plugins = list(
            {n.replace("-", "_") for n in _filter_plugins_from_dists(_get_distributions())}
        )
        locals = [p for p in __modules__ if p != "ape"]
        plugin_modules = tuple([*plugins, *locals])

        for module_name in plugin_modules:
            try:
                module = importlib.import_module(module_name)
                pluggy_manager.register(module)
            except Exception as err:
                if module_name in __modules__:
                    # Always raise core plugin registration errors.
                    raise

                logger.warn_from_exception(err, f"Error loading plugin package '{module_name}'.")

        self.__registered = True

    def _validate_plugin(self, plugin_name: str, plugin_cls) -> Optional[Tuple[str, Tuple]]:
        if valid_impl(plugin_cls):
            return clean_plugin_name(plugin_name), plugin_cls
        else:
            self._warn_not_fully_implemented_error(plugin_cls, plugin_name)
            return None

    def _warn_not_fully_implemented_error(self, results, plugin_name):
        if plugin_name in self._unimplemented_plugins:
            # Already warned
            return

        unimplemented_methods = []

        # Find the best API name to warn about.
        if isinstance(results, (list, tuple)):
            classes = [p for p in results if hasattr(p, "__name__")]
            if classes:
                # Likely only ever a single class in a registration, but just in case.
                api_name = " - ".join([p.__name__ for p in classes if hasattr(p, "__name__")])
                for api_cls in classes:
                    if (
                        abstract_methods := getattr(api_cls, "__abstractmethods__", None)
                    ) and isinstance(abstract_methods, dict):
                        unimplemented_methods.extend(api_cls.__abstractmethods__)

            else:
                # This would only happen if the registration consisted of all primitives.
                api_name = " - ".join(results)

        elif hasattr(results, "__name__"):
            api_name = results.__name__
            if (abstract_methods := getattr(results, "__abstractmethods__", None)) and isinstance(
                abstract_methods, dict
            ):
                unimplemented_methods.extend(results.__abstractmethods__)

        else:
            api_name = results

        message = f"'{api_name}' from '{plugin_name}' is not fully implemented."
        if unimplemented_methods:
            methods_str = ", ".join(unimplemented_methods)
            message = f"{message} Remaining abstract methods: '{methods_str}'."

        logger.warning(message)

        # Record so we don't warn repeatedly
        self._unimplemented_plugins.append(plugin_name)



from typing import Dict, List, Optional

from ape import plugins
from ape.api import PluginConfig


class CustomNetwork(PluginConfig):
    """
    A custom network config.
    """

    name: str
    """Name of the network e.g. mainnet."""

    chain_id: int
    """Chain ID (required)."""

    ecosystem: str
    """The name of the ecosystem."""

    base_ecosystem_plugin: Optional[str] = None
    """The base ecosystem plugin to use, when applicable. Defaults to the default ecosystem."""

    default_provider: str = "geth"
    """The default provider plugin to use. Default is the default node provider."""

    request_header: Dict = {}
    """The HTTP request header."""


class NetworksConfig(PluginConfig):
    custom: List[CustomNetwork] = []


@plugins.register(plugins.Config)
def config_class():
    return NetworksConfig



from fnmatch import fnmatch
from functools import cached_property
from pathlib import Path
from typing import Iterable, List, Set, Union

import click
from click import BadArgumentUsage

from ape.cli.choices import _ACCOUNT_TYPE_FILTER, Alias
from ape.logging import logger
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.os import get_all_files_in_directory, get_full_extension
from ape.utils.validators import _validate_account_alias


def _alias_callback(ctx, param, value):
    return _validate_account_alias(value)


def existing_alias_argument(account_type: _ACCOUNT_TYPE_FILTER = None, **kwargs):
    """
    A ``click.argument`` for an existing account alias.

    Args:
        account_type (Type[:class:`~ape.api.accounts.AccountAPI`], optional):
          If given, limits the type of account the user may choose from.
        **kwargs: click.argument overrides.
    """

    type_ = kwargs.pop("type", Alias(key=account_type))
    return click.argument("alias", type=type_, **kwargs)


def non_existing_alias_argument(**kwargs):
    """
    A ``click.argument`` for an account alias that does not yet exist in ape.

    Args:
        **kwargs: click.argument overrides.
    """

    callback = kwargs.pop("callback", _alias_callback)
    return click.argument("alias", callback=callback, **kwargs)


class _ContractPaths(ManagerAccessMixin):
    """
    Helper callback class for handling CLI-given contract paths.
    """

    def __init__(self, value):
        self.value = value
        self._path_set = set()
        self.missing_compilers = set()
        self.exclude_list = {}

    @classmethod
    def callback(cls, ctx, param, value) -> Set[Path]:
        """
        Use this for click.option / argument callbacks.
        """
        return cls(value).filtered_paths

    @cached_property
    def filtered_paths(self) -> Set[Path]:
        """
        Get the filtered set of paths.
        """
        value = self.value
        contract_paths: Iterable[Path]

        if value and isinstance(value, (list, tuple, set)):
            # Given a single list of paths.
            contract_paths = value

        elif value and isinstance(value, (Path, str)):
            # Given single path.
            contract_paths = (Path(value),)

        elif not value or value == "*":
            # Get all file paths in the project.
            contract_paths = get_all_files_in_directory(self.project_manager.contracts_folder)

        else:
            raise ValueError(f"Unknown contracts-paths value '{value}'.")

        self.lookup(contract_paths)

        # Handle missing compilers.
        if self.missing_compilers:
            # Craft a nice message for all missing compilers.
            missing_ext = ", ".join(sorted(self.missing_compilers))
            message = (
                f"Missing compilers for the following file types: '{missing_ext}'. "
                "Possibly, a compiler plugin is not installed or is "
                "installed but not loading correctly."
            )
            if ".vy" in self.missing_compilers:
                message = f"{message} Is 'ape-vyper' installed?"
            if ".sol" in self.missing_compilers:
                message = f"{message} Is 'ape-solidity' installed?"

            logger.warning(message)

        return self._path_set

    @property
    def exclude_patterns(self) -> List[str]:
        return self.config_manager.get_config("compile").exclude or []

    def do_exclude(self, path: Union[Path, str]) -> bool:
        name = path if isinstance(path, str) else path.name
        if name not in self.exclude_list:
            self.exclude_list[name] = any(fnmatch(name, p) for p in self.exclude_patterns)

        return self.exclude_list[name]

    def compiler_is_unknown(self, path: Union[Path, str]) -> bool:
        path = Path(path)
        if self.do_exclude(path):
            return False

        ext = get_full_extension(path)
        unknown_compiler = ext and ext not in self.compiler_manager.registered_compilers
        if unknown_compiler and ext not in self.missing_compilers:
            self.missing_compilers.add(ext)

        return bool(unknown_compiler)

    def lookup(self, path_iter):
        for path in path_iter:
            path = Path(path)
            if self.do_exclude(path):
                continue

            contracts_folder = self.project_manager.contracts_folder
            if (
                self.project_manager.path / path.name
            ) == contracts_folder or path.name == contracts_folder.name:
                # Was given the path to the contracts folder.
                self.lookup(p for p in self.project_manager.source_paths)

            elif (self.project_manager.path / path).is_dir():
                # Was given sub-dir in the project folder.
                self.lookup(p for p in (self.project_manager.path / path).iterdir())

            elif (contracts_folder / path.name).is_dir():
                # Was given sub-dir in the contracts folder.
                self.lookup(p for p in (contracts_folder / path.name).iterdir())

            elif resolved_path := self.project_manager.lookup_path(path):
                # Check compiler missing.
                if self.compiler_is_unknown(resolved_path):
                    # NOTE: ^ Also tracks.
                    continue

                suffix = get_full_extension(resolved_path)
                if suffix in self.compiler_manager.registered_compilers:
                    # File exists and is compile-able.
                    self._path_set.add(resolved_path)

                elif suffix:
                    raise BadArgumentUsage(f"Source file '{resolved_path.name}' not found.")

            else:
                raise BadArgumentUsage(f"Source file '{path.name}' not found.")


def contract_file_paths_argument():
    """
    A ``click.argument`` representing contract source file paths.
    This argument takes 0-to-many values.

    The return type from the callback is a flattened list of
    source file-paths.
    """
    return click.argument("file_paths", nargs=-1, callback=_ContractPaths.callback)



import json
from pathlib import Path as PathLibPath
from typing import Any, List, Optional

import click
from click import Context, Parameter

from ape.utils import get_all_files_in_directory


class Path(click.Path):
    """
    This class exists to encourage the consistent usage
    of ``pathlib.Path`` for path_type.
    """

    def __init__(self, *args, **kwargs):
        if "path_type" not in kwargs:
            kwargs["path_type"] = PathLibPath

        super().__init__(*args, **kwargs)


# TODO: Delete for 0.8 (list of lists is weird and we
#  are no longer using this).
class AllFilePaths(Path):
    """
    Either all the file paths in the given directory,
    or a list containing only the given file.
    """

    def convert(  # type: ignore[override]
        self, value: Any, param: Optional["Parameter"], ctx: Optional["Context"]
    ) -> List[PathLibPath]:
        path = super().convert(value, param, ctx)
        assert isinstance(path, PathLibPath)  # For mypy

        if not path.is_file() and path.is_absolute():
            # Don't do absolute non-existent paths.
            # Let it resolve elsewhere.
            path = PathLibPath(value)

        return get_all_files_in_directory(path) if path.is_dir() else [path]


class JSON(click.ParamType):
    """
    A type that accepts a raw-JSON str
    and loads it into a dictionary.
    """

    def convert(self, value, param, ctx):
        if not value:
            return {}

        elif isinstance(value, str):
            try:
                return json.loads(value)
            except ValueError as err:
                self.fail(f"Invalid JSON string: {err}", param, ctx)

        return value  # Good already.



import difflib
import re
import sys
import warnings
from gettext import gettext
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import click
import yaml

from ape.cli import ape_cli_context
from ape.exceptions import Abort, ApeException, handle_ape_exception
from ape.logging import logger
from ape.plugins._utils import PluginMetadataList, clean_plugin_name
from ape.utils.basemodel import ManagerAccessMixin

_DIFFLIB_CUT_OFF = 0.6


def display_config(ctx, param, value):
    # NOTE: This is necessary not to interrupt how version or help is intercepted
    if not value or ctx.resilient_parsing:
        return

    click.echo("# Current configuration")

    # NOTE: Using json-mode as yaml.dump requires JSON-like structure.
    model = ManagerAccessMixin.project_manager.config_manager.model_dump(mode="json")

    click.echo(yaml.dump(model))

    ctx.exit()  # NOTE: Must exit to bypass running ApeCLI


class ApeCLI(click.MultiCommand):
    _commands: Optional[Dict] = None
    _CLI_GROUP_NAME = "ape_cli_subcommands"

    def format_commands(self, ctx, formatter) -> None:
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if cmd is None or cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # Allow for 3 times the default spacing.
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            # Split the commands into 3 sections.
            sections: Dict[str, List[Tuple[str, str]]] = {
                "Core": [],
                "Plugin": [],
                "3rd-Party Plugin": [],
            }

            pl_metadata = PluginMetadataList.load(
                ManagerAccessMixin.plugin_manager, include_available=False
            )

            for cli_name, cmd in commands:
                help = cmd.get_short_help_str(limit)
                plugin = pl_metadata.get_plugin(cli_name)
                if not plugin:
                    continue

                if plugin.in_core:
                    sections["Core"].append((cli_name, help))
                elif plugin.is_installed and not plugin.is_third_party:
                    sections["Plugin"].append((cli_name, help))
                else:
                    sections["3rd-Party Plugin"].append((cli_name, help))

            for title, rows in sections.items():
                if not rows:
                    continue

                with formatter.section(gettext(f"{title} Commands")):
                    formatter.write_dl(rows)

    def invoke(self, ctx) -> Any:
        try:
            return super().invoke(ctx)
        except click.UsageError as err:
            self._suggest_cmd(err)
        except ApeException as err:
            path = ctx.obj.project_manager.path

            # NOTE: isinstance check for type-checkers.
            if isinstance(path, Path) and handle_ape_exception(err, (path,)):
                # All exc details already outputted.
                sys.exit(1)
            else:
                raise Abort.from_ape_exception(err) from err

    @staticmethod
    def _suggest_cmd(usage_error):
        if usage_error.message is None:
            raise usage_error

        elif not (match := re.match("No such command '(.*)'.", usage_error.message)):
            raise usage_error

        groups = match.groups()
        if len(groups) < 1:
            raise usage_error

        bad_arg = groups[0]
        suggested_commands = difflib.get_close_matches(
            bad_arg, list(usage_error.ctx.command.commands.keys()), cutoff=_DIFFLIB_CUT_OFF
        )
        if suggested_commands:
            if bad_arg not in suggested_commands:
                usage_error.message = (
                    f"No such command '{bad_arg}'. Did you mean {' or '.join(suggested_commands)}?"
                )

        raise usage_error

    @property
    def commands(self) -> Dict:
        if self._commands:
            return self._commands

        _entry_points = entry_points()
        eps: Iterable
        if select_fn := getattr(_entry_points, "select", None):
            # NOTE: Using getattr because mypy.
            eps = select_fn(group=self._CLI_GROUP_NAME)
        else:
            # Python 3.9. Can remove once we drop support.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                eps = _entry_points.get(self._CLI_GROUP_NAME, [])

        self._commands = {clean_plugin_name(cmd.name): cmd.load for cmd in eps}
        return self._commands

    def list_commands(self, ctx) -> List[str]:
        return list(sorted(self.commands))

    def get_command(self, ctx, name) -> Optional[click.Command]:
        if name in self.commands:
            try:
                return self.commands[name]()
            except Exception as err:
                logger.warn_from_exception(
                    err, f"Unable to load CLI endpoint for plugin 'ape_{name}'"
                )

        # NOTE: don't return anything so Click displays proper error
        return None


@click.command(cls=ApeCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@ape_cli_context()
@click.version_option(message="%(version)s", package_name="eth-ape")
@click.option(
    "--config",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_config,
    help="Show configuration options (using `ape-config.yaml`)",
)
def cli(context):
    _ = context



from pathlib import Path
from typing import Optional, Set

from pydantic import field_validator, model_validator

from ape import plugins
from ape.api import PluginConfig
from ape.utils.misc import SOURCE_EXCLUDE_PATTERNS

DEFAULT_CACHE_FOLDER_NAME = ".cache"  # default relative to contracts/


class Config(PluginConfig):
    include_dependencies: bool = False
    """
    Set to ``True`` to compile dependencies during ``ape compile``.
    Generally, dependencies are not compiled during ``ape compile``
    This is because dependencies may not compile in Ape on their own,
    but you can still reference them in your project's contracts' imports.
    Some projects may be more dependency-based and wish to have the
    contract types always compiled during ``ape compile``, and these projects
    should configure ``include_dependencies`` to be ``True``.
    """

    exclude: Set[str] = set()
    """
    Source exclusion globs across all file types.
    """

    cache_folder: Optional[Path] = None
    """
    Path to contract dependency cache directory (e.g. `contracts/.cache`)
    """

    @property
    def base_path(self) -> Path:
        """The base directory for compilation file path references"""

        # These should be initialized by plugin config loading and pydantic validators before the
        # time this prop is accessed.
        assert self._config_manager is not None
        assert self.cache_folder is not None

        # If the dependency cache folder is configured, to be outside of the contracts dir, we want
        # to use the projects folder to be the base dir for compilation.
        if self._config_manager.contracts_folder not in self.cache_folder.parents:
            return self._config_manager.PROJECT_FOLDER

        # Otherwise, we're defaulting to contracts folder for backwards compatibility. Changing this
        # will cause existing projects to compile to different bytecode.
        return self._config_manager.contracts_folder

    @model_validator(mode="after")
    def validate_cache_folder(self):
        if self._config_manager is None:
            return  # Not enough information to continue at this time

        contracts_folder = self._config_manager.contracts_folder
        project_folder = self._config_manager.PROJECT_FOLDER

        # Set unconfigured default
        if self.cache_folder is None:
            self.cache_folder = contracts_folder / DEFAULT_CACHE_FOLDER_NAME

        # If we get a relative path, assume it's relative to project root (where the config file
        # lives)
        elif not self.cache_folder.is_absolute():
            self.cache_folder = project_folder / self.cache_folder

        # Do not allow escape of the project folder for security and functionality reasons. Paths
        # outside the relative compilation root are not portable and will cause bytecode changes.
        project_resolved = project_folder.resolve()
        cache_resolved = self.cache_folder.resolve()
        if project_resolved not in cache_resolved.parents:
            raise ValueError(
                "cache_folder must be a child of the project directory. "
                f"{project_resolved} not in {cache_resolved}"
            )

    @field_validator("exclude", mode="before")
    @classmethod
    def validate_exclude(cls, value):
        return {*(value or []), *SOURCE_EXCLUDE_PATTERNS}


@plugins.register(plugins.Config)
def config_class():
    return Config



import json
from functools import cached_property
from typing import Collection, Dict, Iterator, List, Optional, Set, Type, Union

import yaml

from ape.api import EcosystemAPI, ProviderAPI, ProviderContextManager
from ape.api.networks import NetworkAPI
from ape.exceptions import EcosystemNotFoundError, NetworkError, NetworkNotFoundError
from ape.managers.base import BaseManager
from ape.utils.basemodel import (
    ExtraAttributesMixin,
    ExtraModelAttributes,
    get_attribute_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import _dict_overlay, log_instead_of_fail
from ape_ethereum.provider import EthereumNodeProvider


class NetworkManager(BaseManager, ExtraAttributesMixin):
    """
    The set of all blockchain network ecosystems registered from the plugin system.
    Typically, you set the provider via the ``--network`` command line option.
    However, use this singleton for more granular access to networks.

    Usage example::

        from ape import networks

        # "networks" is the NetworkManager singleton
        with networks.ethereum.mainnet.use_provider("geth"):
           ...
    """

    _active_provider: Optional[ProviderAPI] = None
    _default_ecosystem_name: Optional[str] = None

    @log_instead_of_fail(default="<NetworkManager>")
    def __repr__(self) -> str:
        provider = self.active_provider
        class_name = NetworkManager.__name__
        content = f"{class_name} active_provider={repr(provider)}" if provider else class_name
        return f"<{content}>"

    @property
    def active_provider(self) -> Optional[ProviderAPI]:
        """
        The currently connected provider if one exists. Otherwise, returns ``None``.
        """

        return self._active_provider

    @active_provider.setter
    def active_provider(self, new_value: ProviderAPI):
        self._active_provider = new_value

    @property
    def network(self) -> NetworkAPI:
        """
        The current network if connected to one.

        Raises:
            :class:`~ape.exceptions.ProviderNotConnectedError`: When there is
              no active provider at runtime.

        Returns:
            :class:`~ape.api.networks.NetworkAPI`
        """
        return self.provider.network

    @property
    def ecosystem(self) -> EcosystemAPI:
        """
        The current ecosystem if connected to one.

        Raises:
            :class:`~ape.exceptions.ProviderNotConnectedError`: When there is
              no active provider at runtime.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """
        return self.network.ecosystem

    def fork(
        self,
        provider_name: Optional[str] = None,
        provider_settings: Optional[Dict] = None,
        block_number: Optional[int] = None,
    ) -> ProviderContextManager:
        """
        Fork the currently connected network.

        Args:
            provider_name (str, optional): The name of the provider to get. Defaults to ``None``.
              When ``None``, returns the default provider.
            provider_settings (dict, optional): Settings to apply to the provider. Defaults to
              ``None``.
            block_number (Optional[int]): Optionally specify the block number you wish to fork.
              Negative block numbers are relative to HEAD. Defaults to the configured fork
              block number or HEAD.

        Returns:
            :class:`~ape.api.networks.ProviderContextManager`
        """
        try:
            forked_network = self.ecosystem.get_network(f"{self.network.name}-fork")
        except NetworkNotFoundError as err:
            raise NetworkError(f"Unable to fork network '{self.network.name}'.") from err

        provider_settings = provider_settings or {}
        fork_settings = {}
        if block_number is not None:
            # Negative block_number means relative to HEAD
            if block_number < 0:
                latest_block_number = self.provider.get_block("latest").number or 0
                block_number = latest_block_number + block_number
                if block_number < 0:
                    # If the block number is still negative, they have forked past genesis.
                    raise NetworkError("Unable to fork past genesis block.")

            # Ensure block_number is set in config for this network
            fork_settings["block_number"] = block_number

        if uri := self.provider.connection_str:
            fork_settings["upstream_provider"] = uri

        _dict_overlay(
            provider_settings,
            {"fork": {self.ecosystem.name: {self.network.name: fork_settings}}},
        )

        shared_kwargs: dict = {"provider_settings": provider_settings, "disconnect_after": True}
        return (
            forked_network.use_provider(provider_name, **shared_kwargs)
            if provider_name
            else forked_network.use_default_provider(**shared_kwargs)
        )

    @property
    def ecosystem_names(self) -> Set[str]:
        """
        The set of all ecosystem names in ``ape``.
        """

        return set(self.ecosystems)

    @property
    def network_names(self) -> Set[str]:
        """
        The set of all network names in ``ape``.
        """

        return {n for e in self.ecosystems.values() for n in e.networks}

    @property
    def provider_names(self) -> Set[str]:
        """
        The set of all provider names in ``ape``.
        """

        return set(
            provider
            for ecosystem in self.ecosystems.values()
            for network in ecosystem.networks.values()
            for provider in network.providers
        )

    @property
    def ecosystems(self) -> Dict[str, EcosystemAPI]:
        """
        All the registered ecosystems in ``ape``, such as ``ethereum``.
        """
        plugin_ecosystems = self._plugin_ecosystems

        # Load config.
        custom_networks: List = self.config_manager.get_config("networks").get("custom", [])
        for custom_network in custom_networks:
            ecosystem_name = custom_network.ecosystem
            if ecosystem_name in plugin_ecosystems:
                # Already included in previous network.
                continue

            base_ecosystem_name = (
                custom_network.get("base_ecosystem_plugin") or self.default_ecosystem_name
            )
            existing_cls = plugin_ecosystems[base_ecosystem_name]
            ecosystem_cls = existing_cls.model_copy(
                update={"name": ecosystem_name}, cache_clear=("_networks_from_plugins",)
            )
            plugin_ecosystems[ecosystem_name] = ecosystem_cls

        return plugin_ecosystems

    @cached_property
    def _plugin_ecosystems(self) -> Dict[str, EcosystemAPI]:
        def to_kwargs(name: str) -> Dict:
            return {
                "name": name,
                "data_folder": self.config_manager.DATA_FOLDER / name,
                "request_header": self.config_manager.REQUEST_HEADER,
            }

        # Load plugins.
        plugins = self.plugin_manager.ecosystems
        return {n: cls(**to_kwargs(n)) for n, cls in plugins}  # type: ignore[operator]

    def create_custom_provider(
        self,
        connection_str: str,
        provider_cls: Type[ProviderAPI] = EthereumNodeProvider,
        provider_name: Optional[str] = None,
    ) -> ProviderAPI:
        """
        Create a custom connection to a URI using the EthereumNodeProvider provider.
        **NOTE**: This provider will assume EVM-like behavior and this is generally not recommended.
        Use plugins when possible!

        Args:
            connection_str (str): The connection string of the node, such as its URI
              when using HTTP.
            provider_cls (Type[:class:`~ape.api.providers.ProviderAPI`]): Defaults to
              :class:`~ape_ethereum.providers.EthereumNodeProvider`.
            provider_name (Optional[str]): The name of the provider. Defaults to best guess.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`: The Geth provider
              implementation that comes with Ape.
        """

        network = self.ethereum.custom_network

        if provider_name is None:
            if issubclass(provider_cls, EthereumNodeProvider):
                name = "geth"

            elif cls_name := getattr(provider_cls, "name", None):
                name = cls_name

            elif cls_name := getattr(provider_cls, "__name__"):
                name = cls_name.lower()

            else:
                # Would be unusual for this to happen though.
                name = "provider"

        else:
            name = provider_name

        provider_settings: Dict = {}
        if connection_str.startswith("https://") or connection_str.startswith("http://"):
            provider_settings["uri"] = connection_str
        elif connection_str.endswith(".ipc"):
            provider_settings["ipc_path"] = connection_str
        else:
            raise NetworkError(f"Scheme for '{connection_str}' not yet supported.")

        return (provider_cls or EthereumNodeProvider)(
            name=name,
            network=network,
            provider_settings=provider_settings,
            data_folder=self.ethereum.data_folder / name,
            request_header=network.request_header,
        )

    def __iter__(self) -> Iterator[str]:
        """
        All the managed ecosystems in ``ape``, as an iterable.

        Returns:
            Iterator[:class:`~ape.api.networks.EcosystemAPI`]
        """
        yield from self.ecosystems

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name="ecosystems",
            attributes=lambda: self.ecosystems,
            include_getitem=True,
        )

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> EcosystemAPI:
        """
        Get an ecosystem via ``.`` access.

        Args:
            attr_name (str): The name of the ecosystem.

        Returns:
            :class:`~ape.api.networks.EcosystemAPI`

        Usage example::

            eth = networks.ethereum
        """
        return get_attribute_with_extras(self, attr_name)

    def get_network_choices(
        self,
        ecosystem_filter: Optional[Union[List[str], str]] = None,
        network_filter: Optional[Union[List[str], str]] = None,
        provider_filter: Optional[Union[List[str], str]] = None,
    ) -> Iterator[str]:
        """
        The set of all possible network choices available as a "network selection"
        e.g. ``--network [ECOSYSTEM:NETWORK:PROVIDER]``.

        Each value is in the form ``ecosystem:network:provider`` and shortened options also
        appear in the list. For example, ``::geth`` would default to ``:ethereum:local:geth``
        and both will be in the returned list. The values come from each
        :class:`~ape.api.providers.ProviderAPI` that is installed.

        Use the CLI command ``ape networks list`` to list all the possible network
        combinations.

        Args:
            ecosystem_filter (Optional[Union[List[str], str]]): Get only the specified ecosystems.
              Defaults to getting all ecosystems.
            network_filter (Optional[Union[List[str], str]]): Get only the specified networks.
              Defaults to getting all networks in ecosystems.
            provider_filter (Optional[Union[List[str], str]]): Get only the specified providers.
              Defaults to getting all providers in networks.

        Returns:
            Iterator[str]: An iterator over all the network-choice possibilities.
        """

        ecosystem_filter = _validate_filter(ecosystem_filter, self.ecosystem_names)
        network_filter = _validate_filter(network_filter, self.network_names)
        provider_filter = _validate_filter(provider_filter, self.provider_names)

        ecosystem_items = self.ecosystems
        if ecosystem_filter:
            ecosystem_items = {n: e for n, e in ecosystem_items.items() if n in ecosystem_filter}

        for ecosystem_name, ecosystem in ecosystem_items.items():
            network_items = ecosystem.networks
            if network_filter:
                network_items = {n: net for n, net in network_items.items() if n in network_filter}

            if not network_items:
                continue

            ecosystem_has_providers = False
            for network_name, network in network_items.items():
                providers = network.providers
                if provider_filter:
                    providers = [n for n in providers if n in provider_filter]

                network_has_providers = len(providers) > 0
                if not ecosystem_has_providers:
                    # Only check if we still haven't found any
                    ecosystem_has_providers = network_has_providers

                if not network_has_providers:
                    continue

                for provider_name in providers:
                    if (
                        ecosystem_name == self.default_ecosystem.name
                        and network_name == ecosystem.default_network_name
                    ):
                        yield f"::{provider_name}"

                    if ecosystem_name == self.default_ecosystem.name:
                        yield f":{network_name}:{provider_name}"

                    if network_name == ecosystem.default_network_name:
                        yield f"{ecosystem_name}::{provider_name}"

                    # Always include the full path as an option.
                    yield f"{ecosystem_name}:{network_name}:{provider_name}"

                # Providers were yielded if we reached this point.
                if ecosystem_name == self.default_ecosystem.name:
                    yield f":{network_name}"

                yield f"{ecosystem_name}:{network_name}"

            if ecosystem_has_providers:
                yield ecosystem_name

    def get_ecosystem(self, ecosystem_name: str) -> EcosystemAPI:
        """
        Get the ecosystem for the given name.

        Args:
            ecosystem_name (str): The name of the ecosystem to get.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the ecosystem is not found.

        Returns:
            :class:`~ape.api.networks.EcosystemAPI`
        """

        if ecosystem_name not in self.ecosystem_names:
            raise EcosystemNotFoundError(ecosystem_name, options=self.ecosystem_names)

        return self.ecosystems[ecosystem_name]

    def get_provider_from_choice(
        self,
        network_choice: Optional[str] = None,
        provider_settings: Optional[Dict] = None,
    ) -> ProviderAPI:
        """
        Get a :class:`~ape.api.providers.ProviderAPI` from a network choice.
        A network choice is any value returned from
        :meth:`~ape.managers.networks.NetworkManager.get_network_choices`. Use the
        CLI command ``ape networks list`` to list all the possible network
        combinations.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given network choice does not
              match any known network.

        Args:
            network_choice (str, optional): The network choice
              (see :meth:`~ape.managers.networks.NetworkManager.get_network_choices`).
              Defaults to the default ecosystem, network, and provider combination.
            provider_settings (dict, optional): Settings for the provider. Defaults to None.

        Returns:
            :class:`~ape.api.providers.ProviderAPI`
        """

        if network_choice is None:
            default_network = self.default_ecosystem.default_network
            return default_network.get_provider(provider_settings=provider_settings)

        elif _is_custom_network(network_choice):
            # Custom network w/o ecosystem & network spec.
            return self.create_custom_provider(network_choice)

        selections = network_choice.split(":")

        # NOTE: Handle case when URI is passed e.g. "http://..."
        if len(selections) > 3:
            provider_value = ":".join(selections[2:])
            selections[2] = provider_value
            selections = selections[:3]
            if _is_custom_network(provider_value):
                selections[1] = selections[1] or "custom"

        if selections == network_choice or len(selections) == 1:
            # Either split didn't work (in which case it matches the start)
            # or there was nothing after the ``:`` (e.g. "ethereum:")
            ecosystem = self.get_ecosystem(selections[0] or self.default_ecosystem.name)
            # By default, the "local" network should be specified for
            # any ecosystem (this should not correspond to a production chain)
            default_network = ecosystem.default_network
            return default_network.get_provider(provider_settings=provider_settings)

        elif len(selections) == 2:
            # Only ecosystem and network were specified, not provider
            ecosystem_name, network_name = selections
            ecosystem = self.get_ecosystem(ecosystem_name or self.default_ecosystem.name)
            network = ecosystem.get_network(network_name or ecosystem.default_network_name)
            return network.get_provider(provider_settings=provider_settings)

        elif len(selections) == 3:
            # Everything is specified, use specified provider for ecosystem and network
            ecosystem_name, network_name, provider_name = selections
            ecosystem = (
                self.get_ecosystem(ecosystem_name) if ecosystem_name else self.default_ecosystem
            )
            network = ecosystem.get_network(network_name or ecosystem.default_network_name)
            return network.get_provider(
                provider_name=provider_name, provider_settings=provider_settings
            )

        else:
            # NOTE: Might be unreachable
            raise NetworkError("Invalid network selection.")

    def parse_network_choice(
        self,
        network_choice: Optional[str] = None,
        provider_settings: Optional[Dict] = None,
        disconnect_after: bool = False,
        disconnect_on_exit: bool = True,
    ) -> ProviderContextManager:
        """
        Parse a network choice into a context manager for managing a temporary
        connection to a provider. See
        :meth:`~ape.managers.networks.NetworkManager.get_network_choices` for all
        available choices (or use CLI command ``ape networks list``).

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given network choice does not
              match any known network.

        Args:
            network_choice (str, optional): The network choice
              (see :meth:`~ape.managers.networks.NetworkManager.get_network_choices`).
              Defaults to the default ecosystem, network, and provider combination.
            provider_settings (dict, optional): Settings for the provider. Defaults to None.
            disconnect_after (bool): Set to True to terminate the connection completely
              at the end of context. NOTE: May only work if the network was also started
              from this session.
            disconnect_on_exit (bool): Whether to disconnect on the exit of the python
              session. Defaults to ``True``.

        Returns:
            :class:`~api.api.networks.ProviderContextManager`
        """

        provider = self.get_provider_from_choice(
            network_choice=network_choice, provider_settings=provider_settings
        )
        return ProviderContextManager(
            provider=provider,
            disconnect_after=disconnect_after,
            disconnect_on_exit=disconnect_on_exit,
        )

    @property
    def default_ecosystem_name(self) -> str:
        if name := self._default_ecosystem_name:
            return name

        return self.config_manager.default_ecosystem or "ethereum"

    @property
    def default_ecosystem(self) -> EcosystemAPI:
        """
        The default ecosystem. Call
        :meth:`~ape.managers.networks.NetworkManager.set_default_ecosystem` to
        change the default ecosystem. If a default is not set and there is
        only a single ecosystem installed, such as Ethereum, then get
        that ecosystem.
        """

        return self.ecosystems[self.default_ecosystem_name]

    def set_default_ecosystem(self, ecosystem_name: str):
        """
        Change the default ecosystem.

        Raises:
            :class:`~ape.exceptions.NetworkError`: When the given ecosystem name is unknown.

        Args:
            ecosystem_name (str): The name of the ecosystem to set
              as the default.
        """

        if ecosystem_name in self.ecosystem_names:
            self._default_ecosystem_name = ecosystem_name

        else:
            raise EcosystemNotFoundError(ecosystem_name, options=self.ecosystem_names)

    @property
    def network_data(self) -> Dict:
        """
        Get a dictionary containing data about networks in the ecosystem.

        **NOTE**: The keys are added in an opinionated order for nicely
        translating into ``yaml``.

        Returns:
            dict
        """
        return self.get_network_data()

    def get_network_data(
        self,
        ecosystem_filter: Optional[Collection[str]] = None,
        network_filter: Optional[Collection[str]] = None,
        provider_filter: Optional[Collection[str]] = None,
    ):
        data: Dict = {"ecosystems": []}

        for ecosystem_name in self:
            if ecosystem_filter and ecosystem_name not in ecosystem_filter:
                continue

            ecosystem_data = self._get_ecosystem_data(
                ecosystem_name, network_filter=network_filter, provider_filter=provider_filter
            )
            data["ecosystems"].append(ecosystem_data)

        return data

    def _get_ecosystem_data(
        self,
        ecosystem_name: str,
        network_filter: Optional[Collection[str]] = None,
        provider_filter: Optional[Collection[str]] = None,
    ) -> Dict:
        ecosystem = self[ecosystem_name]
        ecosystem_data: Dict = {"name": str(ecosystem_name)}

        # Only add isDefault key when True
        if ecosystem_name == self.default_ecosystem.name:
            ecosystem_data["isDefault"] = True

        ecosystem_data["networks"] = []
        networks = getattr(self, ecosystem_name).networks

        for network_name in networks:
            if network_filter and network_name not in network_filter:
                continue

            network_data = ecosystem.get_network_data(network_name, provider_filter=provider_filter)
            ecosystem_data["networks"].append(network_data)

        return ecosystem_data

    @property
    # TODO: Remove in 0.7
    def networks_yaml(self) -> str:
        """
        Get a ``yaml`` ``str`` representing all the networks
        in all the ecosystems.
        **NOTE**: Deprecated.

        View the result via CLI command ``ape networks list --format yaml``.

        Returns:
            str
        """

        data = self.network_data
        if not isinstance(data, dict):
            raise TypeError(
                f"Unexpected network data type: {type(data)}. "
                f"Expecting dict. YAML dump will fail."
            )

        try:
            return yaml.dump(data, sort_keys=True)
        except ValueError as err:
            try:
                data_str = json.dumps(data)
            except Exception:
                data_str = str(data)

            raise NetworkError(
                f"Network data did not dump to YAML: {data_str}\nActual err: {err}"
            ) from err


def _validate_filter(arg: Optional[Union[List[str], str]], options: Set[str]):
    filters = arg or []

    if isinstance(filters, str):
        filters = [filters]

    for _filter in filters:
        if _filter not in options:
            raise NetworkError(f"Unknown option '{_filter}'.")

    return filters


def _is_custom_network(value: str) -> bool:
    return (
        value.startswith("http://")
        or value.startswith("https://")
        or value.startswith("ws://")
        or value.startswith("wss://")
        or (value.endswith(".ipc") and ":" not in value)
    )



"""Utilities to extract debug log data from Vyper's print() and the Hardhat Solidity console.log
calls.

There are two different implementations of the console.log functionality. Vyper's print() and
Hardhat's console.sol contract. Both send staticcalls to a specific non-existent contract that we
can read from the call tree.

Vyper encodes these calls as `log(string,bytes)`. Where the first argument is an ABI signature
(e.g. "(uint256,uint8,string)") and the second arg is the printed data. This allows for a simple ABI
and very dynamic output.

Hardhat's console.sol has a strict ABI (unless patched :-/) that mixes and matches all the basic
datatypes to overload all calls to `log()`.

References
----------
- Vyper `print()`: https://docs.vyperlang.org/en/latest/built-in-functions.html#print
- Hardhat's `console.sol`: https://github.com/NomicFoundation/hardhat/blob/e6a42cf330863a6c64624baedd82fc6c5218a20a/packages/hardhat-core/console.sol#L4  # noqa: E501
- Discussion on dynamic ABI encoding (Vyper-style) for log calls: https://github.com/NomicFoundation/hardhat/issues/2666  # noqa: E501
"""

from typing import Any, Iterable, Tuple, cast

from eth_abi import decode
from eth_typing import ChecksumAddress
from eth_utils import decode_hex
from ethpm_types import ContractType, MethodABI
from typing_extensions import TypeGuard

import ape
from ape.types import CallTreeNode

from ._console_log_abi import CONSOLE_LOG_ABI

CONSOLE_CONTRACT_ID = cast(ChecksumAddress, "0x000000000000000000636F6e736F6c652e6c6f67")
VYPER_PRINT_METHOD_ID = "0x23cdd8e8"  # log(string,bytes)

console_contract = ContractType(abi=CONSOLE_LOG_ABI, contractName="console")


def is_console_log(call: Any) -> TypeGuard[CallTreeNode]:
    """Determine if a call is a starndard console.log() call"""
    return (
        isinstance(call, CallTreeNode)
        and call.contract_id == CONSOLE_CONTRACT_ID
        and call.method_id in console_contract.identifier_lookup
    )


def is_vyper_print(call: Any) -> TypeGuard[CallTreeNode]:
    """Determine if a call is a starndard Vyper print() call"""
    if (
        isinstance(call, CallTreeNode)
        and call.contract_id == CONSOLE_CONTRACT_ID
        and call.method_id == VYPER_PRINT_METHOD_ID
        and isinstance(call.inputs, str)
    ):
        bcalldata = decode_hex(call.inputs)
        schema, _ = decode(["string", "bytes"], bcalldata)
        try:
            # Now we look at the first arg to try and determine if it's an ABI signature
            first_type = schema.strip("()").split(",")[0]
            # TODO: Tighten this up.  This is not entirely accurate, but should mostly get us there.
            if (
                first_type.startswith("uint")
                or first_type.startswith("int")
                or first_type.startswith("bytes")
                or first_type == "string"
            ):
                return True
        except IndexError:
            # Empty string as first arg?
            pass
    return False


def console_log(method_abi: MethodABI, calldata: str) -> Tuple[Any]:
    """Return logged data for console.log() calls"""
    bcalldata = decode_hex(calldata)
    data = ape.networks.ethereum.decode_calldata(method_abi, bcalldata)
    return tuple(data.values())


def vyper_print(calldata: str) -> Tuple[Any]:
    """Return logged data for print() calls"""
    bcalldata = decode_hex(calldata)
    schema, payload = decode(["string", "bytes"], bcalldata)
    data = decode(schema.strip("()").split(","), payload)
    return tuple(data)


def extract_debug_logs(call_tree: CallTreeNode) -> Iterable[Tuple[Any]]:
    """Filter calls to console.log() and print() from a transactions call tree"""
    for call in call_tree.calls:
        if is_vyper_print(call) and call.inputs is not None:
            yield vyper_print(call.inputs)
        elif is_console_log(call) and call.inputs is not None:
            assert call.method_id is not None  # is_console_log check already checked
            method_abi = console_contract.identifier_lookup.get(call.method_id)
            if isinstance(method_abi, MethodABI):
                yield console_log(method_abi, call.inputs)
        elif call.calls is not None:
            yield from extract_debug_logs(call)



import sys
from enum import Enum, IntEnum
from functools import cached_property
from typing import IO, Any, Dict, List, Optional, Tuple, Union

from eth_abi import decode
from eth_account import Account as EthAccount
from eth_account._utils.legacy_transactions import (
    encode_transaction,
    serializable_unsigned_transaction_from_dict,
)
from eth_pydantic_types import HexBytes
from eth_utils import decode_hex, encode_hex, keccak, to_hex, to_int
from ethpm_types import ContractType
from ethpm_types.abi import EventABI, MethodABI
from pydantic import BaseModel, Field, field_validator, model_validator

from ape.api import ReceiptAPI, TransactionAPI
from ape.contracts import ContractEvent
from ape.exceptions import APINotImplementedError, OutOfGasError, SignatureError, TransactionError
from ape.logging import logger
from ape.types import AddressType, CallTreeNode, ContractLog, ContractLogContainer, SourceTraceback
from ape.utils import ZERO_ADDRESS
from ape_ethereum._print import extract_debug_logs


class TransactionStatusEnum(IntEnum):
    """
    An ``Enum`` class representing the status of a transaction.
    """

    FAILING = 0
    """The transaction has failed or is in the process of failing."""

    NO_ERROR = 1
    """
    The transaction is successful and is confirmed or is in the process
    of getting confirmed.
    """


class TransactionType(Enum):
    """
    Transaction enumerable type constants defined by
    `EIP-2718 <https://eips.ethereum.org/EIPS/eip-2718>`__.
    """

    STATIC = 0
    ACCESS_LIST = 1  # EIP-2930
    DYNAMIC = 2  # EIP-1559
    SHARED_BLOB = 3  # EIP-4844


class AccessList(BaseModel):
    address: AddressType
    storage_keys: List[HexBytes] = Field(default_factory=list, alias="storageKeys")


class BaseTransaction(TransactionAPI):
    def serialize_transaction(self) -> bytes:
        if not self.signature:
            message = "The transaction is not signed."
            if not self.sender:
                message = (
                    f"{message} "
                    "Did you forget to add the `sender=` kwarg to the transaction function call?"
                )

            raise SignatureError(message)

        txn_data = self.model_dump(by_alias=True, exclude={"sender", "type"})

        # This messes up the signature
        if txn_data.get("to") == ZERO_ADDRESS:
            del txn_data["to"]

        # Adjust bytes in the access list if necessary.
        if "accessList" in txn_data:
            adjusted_access_list = []

            for item in txn_data["accessList"]:
                adjusted_item = {**item}
                storage_keys_corrected = [HexBytes(k).hex() for k in item.get("storageKeys", [])]

                if storage_keys_corrected:
                    adjusted_item["storageKeys"] = storage_keys_corrected

                adjusted_access_list.append(adjusted_item)

            txn_data["accessList"] = adjusted_access_list

        unsigned_txn = serializable_unsigned_transaction_from_dict(txn_data)
        signature = (self.signature.v, to_int(self.signature.r), to_int(self.signature.s))
        signed_txn = encode_transaction(unsigned_txn, signature)
        impersonated_accounts = self.account_manager.test_accounts._impersonated_accounts

        # If this is a real sender (not impersonated), verify its signature.
        if self.sender and self.sender not in impersonated_accounts:
            recovered_signer = EthAccount.recover_transaction(signed_txn)
            if recovered_signer != self.sender:
                raise SignatureError(
                    f"Recovered signer '{recovered_signer}' doesn't match sender {self.sender}!"
                )

        return signed_txn

    @property
    def txn_hash(self) -> HexBytes:
        txn_bytes = self.serialize_transaction()
        return HexBytes(keccak(txn_bytes))


class StaticFeeTransaction(BaseTransaction):
    """
    Transactions that are pre-EIP-1559 and use the ``gasPrice`` field.
    """

    gas_price: Optional[int] = Field(None, alias="gasPrice")
    max_priority_fee: Optional[int] = Field(None, exclude=True)  # type: ignore
    type: int = Field(TransactionType.STATIC.value, exclude=True)
    max_fee: Optional[int] = Field(None, exclude=True)  # type: ignore

    @model_validator(mode="before")
    @classmethod
    def calculate_read_only_max_fee(cls, values) -> Dict:
        # NOTE: Work-around, Pydantic doesn't handle calculated fields well.
        values["max_fee"] = cls.conversion_manager.convert(
            values.get("gas_limit", 0), int
        ) * cls.conversion_manager.convert(values.get("gas_price", 0), int)
        return values


class AccessListTransaction(StaticFeeTransaction):
    """
    `EIP-2930 <https://eips.ethereum.org/EIPS/eip-2930>`__
    transactions are similar to legacy transaction with an added access list functionality.
    """

    gas_price: Optional[int] = Field(None, alias="gasPrice")
    type: int = Field(TransactionType.ACCESS_LIST.value)
    access_list: List[AccessList] = Field(default_factory=list, alias="accessList")

    @field_validator("type")
    @classmethod
    def check_type(cls, value):
        return value.value if isinstance(value, TransactionType) else value


class DynamicFeeTransaction(BaseTransaction):
    """
    Transactions that are post-EIP-1559 and use the ``maxFeePerGas``
    and ``maxPriorityFeePerGas`` fields.
    """

    max_priority_fee: Optional[int] = Field(None, alias="maxPriorityFeePerGas")  # type: ignore
    max_fee: Optional[int] = Field(None, alias="maxFeePerGas")  # type: ignore
    type: int = Field(TransactionType.DYNAMIC.value)
    access_list: List[AccessList] = Field(default_factory=list, alias="accessList")

    @field_validator("type")
    @classmethod
    def check_type(cls, value):
        return value.value if isinstance(value, TransactionType) else value


class SharedBlobTransaction(DynamicFeeTransaction):
    """
    `EIP-4844 <https://eips.ethereum.org/EIPS/eip-4844>`__ transactions.
    """

    max_fee_per_blob_gas: int = Field(0, alias="maxFeePerBlobGas")
    blob_versioned_hashes: List[HexBytes] = Field([], alias="blobVersionedHashes")

    """
    Overridden because EIP-4844 states it cannot be nil.
    """
    receiver: AddressType = Field(ZERO_ADDRESS, alias="to")

    @field_validator("max_fee_per_blob_gas", mode="before")
    @classmethod
    def hex_to_int(cls, value):
        return value if isinstance(value, int) else int(HexBytes(value).hex(), 16)


class Receipt(ReceiptAPI):
    gas_limit: int
    gas_price: int
    gas_used: int

    @property
    def ran_out_of_gas(self) -> bool:
        return (
            self.status == TransactionStatusEnum.FAILING.value and self.gas_used == self.gas_limit
        )

    @property
    def total_fees_paid(self) -> int:
        """
        The total amount of fees paid for the transaction.
        """
        return self.gas_used * self.gas_price

    @property
    def failed(self) -> bool:
        return self.status != TransactionStatusEnum.NO_ERROR

    @cached_property
    def call_tree(self) -> Optional[CallTreeNode]:
        return self.provider.get_call_tree(self.txn_hash)

    @cached_property
    def debug_logs_typed(self) -> List[Tuple[Any]]:
        """
        Extract messages to console outputted by contracts via print() or console.log() statements
        """

        try:
            self.call_tree
        # Some providers do not implement this, so skip
        except APINotImplementedError:
            logger.debug("Call tree not available, skipping debug log extraction")
            return list()

        # If the call tree is not available, no logs are available
        if self.call_tree is None:
            return list()

        return list(extract_debug_logs(self.call_tree))

    @cached_property
    def contract_type(self) -> Optional[ContractType]:
        if address := (self.receiver or self.contract_address):
            return self.chain_manager.contracts.get(address)

        return None

    @cached_property
    def method_called(self) -> Optional[MethodABI]:
        if not self.contract_type:
            return None

        method_id = self.data[:4]
        if method_id not in self.contract_type.methods:
            return None

        return self.contract_type.methods[method_id]

    @cached_property
    def source_traceback(self) -> SourceTraceback:
        if contract_type := self.contract_type:
            try:
                return SourceTraceback.create(contract_type, self.trace, HexBytes(self.data))
            except Exception as err:
                # Failing to get a traceback should not halt an Ape application.
                # Sometimes, a node crashes and we are left with nothing.
                logger.error(f"Problem retrieving traceback: {err}")

        return SourceTraceback.model_validate([])

    def raise_for_status(self):
        if self.gas_limit is not None and self.ran_out_of_gas:
            raise OutOfGasError(txn=self)

        elif self.status != TransactionStatusEnum.NO_ERROR:
            txn_hash = HexBytes(self.txn_hash).hex()
            raise TransactionError(f"Transaction '{txn_hash}' failed.", txn=self)

    def show_trace(self, verbose: bool = False, file: IO[str] = sys.stdout):
        if not (call_tree := self.call_tree):
            return

        call_tree.enrich(use_symbol_for_tokens=True)
        revert_message = None

        if call_tree.failed:
            default_message = "reverted without message"
            returndata = HexBytes(call_tree.raw["returndata"])
            if to_hex(returndata).startswith(
                "0x08c379a00000000000000000000000000000000000000000000000000000000000000020"
            ):
                # Extra revert-message
                decoded_result = decode(("string",), returndata[4:])
                if len(decoded_result) == 1:
                    revert_message = f'reverted with message: "{decoded_result[0]}"'
                else:
                    revert_message = default_message

            elif address := (self.receiver or self.contract_address):
                # Try to enrich revert error using ABI.
                if provider := self.network_manager.active_provider:
                    ecosystem = provider.network.ecosystem
                else:
                    # Default to Ethereum.
                    ecosystem = self.network_manager.ethereum

                try:
                    instance = ecosystem.decode_custom_error(returndata, address)
                except NotImplementedError:
                    pass
                else:
                    revert_message = repr(instance)

        self.chain_manager._reports.show_trace(
            call_tree,
            sender=self.sender,
            transaction_hash=self.txn_hash,
            revert_message=revert_message,
            verbose=verbose,
            file=file,
        )

    def show_gas_report(self, file: IO[str] = sys.stdout):
        if not (call_tree := self.call_tree):
            return

        call_tree.enrich()

        # Enrich transfers.
        if (
            call_tree.contract_id.startswith("Transferring ")
            and self.receiver is not None
            and self.receiver in self.account_manager
        ):
            receiver_id = self.account_manager[self.receiver].alias or self.receiver
            call_tree.method_id = f"to:{receiver_id}"

        elif call_tree.contract_id.startswith("Transferring "):
            call_tree.method_id = f"to:{self.receiver}"

        self.chain_manager._reports.show_gas(call_tree, file=file)

    def show_source_traceback(self, file: IO[str] = sys.stdout):
        self.chain_manager._reports.show_source_traceback(
            self.source_traceback, file=file, failing=self.failed
        )

    def decode_logs(
        self,
        abi: Optional[
            Union[List[Union[EventABI, "ContractEvent"]], Union[EventABI, "ContractEvent"]]
        ] = None,
    ) -> ContractLogContainer:
        if not self.logs:
            # Short circuit.
            return ContractLogContainer([])

        elif abi is not None:
            if not isinstance(abi, (list, tuple)):
                abi = [abi]

            event_abis: List[EventABI] = [a.abi if not isinstance(a, EventABI) else a for a in abi]
            return ContractLogContainer(
                self.provider.network.ecosystem.decode_logs(self.logs, *event_abis)
            )

        else:
            # If ABI is not provided, decode all events
            addresses = {x["address"] for x in self.logs}
            contract_types = self.chain_manager.contracts.get_multiple(addresses)
            # address → selector → abi
            selectors = {
                address: {encode_hex(keccak(text=abi.selector)): abi for abi in contract.events}
                for address, contract in contract_types.items()
            }

            def get_default_log(
                _log: Dict, logs: ContractLogContainer, evt_name: Optional[str] = None
            ) -> ContractLog:
                log_index = _log.get("logIndex", logs[-1].log_index + 1 if logs else 0)

                # NOTE: Happens when decoding fails.
                evt_name = evt_name or f"UnknownLog_WithIndex_{log_index}"

                return ContractLog(
                    block_hash=self.block.hash,
                    block_number=self.block_number,
                    event_arguments={"root": _log["data"]},
                    event_name=f"<{evt_name}>",
                    log_index=log_index,
                    transaction_hash=self.txn_hash,
                    transaction_index=logs[-1].transaction_index if logs else None,
                )

            decoded_logs: ContractLogContainer = ContractLogContainer()
            for log in self.logs:
                if contract_address := log.get("address"):
                    if contract_address in selectors and (topics := log.get("topics")):
                        selector = encode_hex(topics[0])
                        if selector in selectors[contract_address]:
                            event_abi = selectors[contract_address][selector]
                            decoded_logs.extend(
                                self.provider.network.ecosystem.decode_logs([log], event_abi)
                            )

                        elif library_log := self._decode_ds_note(log):
                            decoded_logs.append(library_log)

                        else:
                            # Search for selector in other spots:
                            name = f"UnknownLogWithSelector_{selector}"
                            obj = get_default_log(log, decoded_logs, evt_name=name)
                            decoded_logs.append(obj)

                    elif library_log := self._decode_ds_note(log):
                        decoded_logs.append(library_log)

                    else:
                        name = f"UnknownLogAtAddress_{contract_address}"
                        index = log.get("logIndex")
                        if index is not None:
                            name = f"{name}_AndLogIndex_{index}"

                        obj = get_default_log(log, decoded_logs, evt_name=name)
                        decoded_logs.append(obj)

                elif library_log := self._decode_ds_note(log):
                    decoded_logs.append(library_log)

                else:
                    obj = get_default_log(log, decoded_logs)
                    decoded_logs.append(obj)

            return decoded_logs

    def _decode_ds_note(self, log: Dict) -> Optional[ContractLog]:
        # The first topic encodes the function selector
        selector, tail = log["topics"][0][:4], log["topics"][0][4:]
        if sum(tail):
            # non-zero bytes found after selector
            return None

        if not (contract_type := self.chain_manager.contracts.get(log["address"])):
            # contract type for {log['address']} not found
            return None

        try:
            method_abi = contract_type.mutable_methods[selector]
        except KeyError:
            #  selector {selector.hex()} not found in {log['address']}
            return None

        # ds-note data field uses either (uint256,bytes) or (bytes) encoding
        # instead of guessing, assume the payload begins right after the selector
        data = decode_hex(log["data"]) if isinstance(log["data"], str) else log["data"]
        input_types = [i.canonical_type for i in method_abi.inputs]
        start_index = data.index(selector) + 4
        values = decode(input_types, data[start_index:])
        address = self.provider.network.ecosystem.decode_address(log["address"])

        return ContractLog(
            block_hash=log["blockHash"],
            block_number=log["blockNumber"],
            contract_address=address,
            event_arguments={i.name: value for i, value in zip(method_abi.inputs, values)},
            event_name=method_abi.name,
            log_index=log["logIndex"],
            transaction_hash=log["transactionHash"],
            transaction_index=log["transactionIndex"],
        )


class SharedBlobReceipt(Receipt):
    """
    An `EIP-4844 <https://eips.ethereum.org/EIPS/eip-4844#blob-transaction>`__"
    blob transaction.
    """

    blob_gas_used: int
    """
    The total amount of blob gas consumed by the transactions within the block.
    """

    blob_gas_price: int
    """
    The blob-gas price, independent from regular gas price.
    """

    @field_validator("blob_gas_used", "blob_gas_price", mode="before")
    @classmethod
    def validate_hex(cls, value):
        return cls._hex_to_int(value or 0)

    @classmethod
    def _hex_to_int(cls, value) -> int:
        return value if isinstance(value, int) else int(HexBytes(value).hex(), 16)



import os
import shutil
from pathlib import Path
from typing import Dict

import pytest
from pydantic import ValidationError

from ape.exceptions import ProjectError
from ape.managers.project.dependency import GithubDependency, LocalDependency, NpmDependency
from ape.utils import create_tempdir


@pytest.fixture
def oz_dependencies_config():
    def _create_oz_dependency(version: str) -> Dict:
        return {
            "name": "OpenZeppelin",
            "version": version,
            "github": "OpenZeppelin/openzeppelin-contracts",
        }

    return {"dependencies": [_create_oz_dependency("3.1.0"), _create_oz_dependency("4.4.2")]}


@pytest.fixture
def local_dependency(project_with_dependency_config):
    return project_with_dependency_config.dependencies["testdependency"]["local"]


@pytest.fixture
def project_with_downloaded_dependencies(temp_config, config, oz_dependencies_config):
    manifests_directory = Path(__file__).parent / "data" / "manifests"
    oz_manifests = manifests_directory / "OpenZeppelin"
    oz_manifests_dest = config.packages_folder / "OpenZeppelin"

    if oz_manifests_dest.is_dir():
        shutil.rmtree(oz_manifests_dest)

    shutil.copytree(oz_manifests, oz_manifests_dest)
    with temp_config(oz_dependencies_config) as project:
        yield project


def test_two_dependencies_with_same_name(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_310 = project_with_downloaded_dependencies.dependencies[name]["3.1.0"]
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    base_uri = "https://github.com/OpenZeppelin/openzeppelin-contracts/releases/tag"

    assert oz_310.version == "3.1.0"
    assert oz_310.name == name
    assert str(oz_310.uri) == f"{base_uri}/v3.1.0"
    assert oz_442.version == "4.4.2"
    assert oz_442.name == name
    assert str(oz_442.uri) == f"{base_uri}/v4.4.2"


def test_dependency_contracts_folder(config, local_dependency):
    """
    The local dependency fixture uses a longer contracts folder path.
    This test ensures that the contracts folder field is honored, specifically
    In the case when it contains sub-paths.
    """
    actual = local_dependency.contracts_folder
    assert actual == "source/v0.1"


def test_local_dependency(local_dependency, config):
    assert local_dependency.name == "testdependency"
    assert local_dependency.version_id == "local"
    assert str(local_dependency.uri).startswith("file://")


def test_access_dependency_contracts(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    contract = oz_442.AccessControl
    assert contract.contract_type.name == "AccessControl"


@pytest.mark.parametrize("ref", ("main", "v1.0.0", "1.0.0"))
def test_dependency_using_reference(ref, recwarn, dependency_manager):
    dependency_config = {
        "github": "apeworx/testfoobartest",
        "name": "foobar",
        "ref": ref,
    }
    dependency = dependency_manager.decode_dependency(dependency_config)
    _ = dependency.cached_manifest
    assert dependency.version is None
    assert dependency.ref == ref
    assert str(dependency.uri) == f"https://github.com/apeworx/testfoobartest/tree/{ref}"

    # Tests against bug where we tried creating version objects out of branch names,
    # thus causing warnings to show in `ape test` runs.
    assert DeprecationWarning not in [w.category for w in recwarn.list]


def test_npm_dependency(mock_home_directory):
    name = "@gnosis.pm"
    package = "safe-singleton-factory"
    version = "1.0.0"
    dependency = NpmDependency(name=package, npm=f"{name}/{package}", version=version)
    with create_tempdir() as temp_dir:
        os.chdir(str(temp_dir))

        # Test with both local and global node modules install.
        for base in (temp_dir, mock_home_directory):
            package_folder = base / "node_modules" / name / package
            contracts_folder = package_folder / "contracts"
            contracts_folder.mkdir(parents=True)
            package_json = package_folder / "package.json"
            package_json.write_text(f'{{"version": "{version}"}}')
            file = contracts_folder / "contract.json"
            source_content = '{"abi": []}'
            file.write_text(source_content)

            manifest = dependency.extract_manifest(use_cache=False)

            assert manifest.sources
            assert str(manifest.sources["contract.json"].content) == f"{source_content}\n"

            shutil.rmtree(package_folder)


def test_decode_with_config_override(dependency_manager, project):
    settings = {".json": {"evm_version": "paris"}}
    path = "__test_path__"
    base_path = project.path / path
    contracts_path = base_path / "contracts"
    contracts_path.mkdir(parents=True)
    (contracts_path / "contract.json").write_text('{"abi": []}')

    data = {"name": "FooBar", "local": path, "config_override": settings}
    dependency = dependency_manager.decode_dependency(data)
    assert dependency.config_override == settings


def test_compile(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    # NOTE: the test data is pre-compiled because the ape-solidity plugin is required.
    actual = oz_442.compile()
    assert len(actual.contract_types) > 0


def test_compile_with_config_override(dependency_manager, project):
    # NOTE: It is important that `contracts_folder` is present in settings
    #  for this test to test against a previous bug where we got multiple values.
    override = {"contracts_folder": "src"}
    path = "__test_path__"
    contracts_path = project.path / path / "src"
    contracts_path.mkdir(exist_ok=True, parents=True)
    (contracts_path / "contract.json").write_text('{"abi": []}')
    data = {"name": "FooBar", "local": path, "config_override": override}
    dependency = dependency_manager.decode_dependency(data)

    actual = dependency.compile()
    assert len(actual.contract_types) > 0


def test_github_dependency_ref_or_version_is_required():
    expected = r"GitHub dependency must have either ref or version specified"
    with pytest.raises(ValidationError, match=expected):
        _ = GithubDependency(name="foo", github="asdf")


def test_dependency_missing_sources():
    """
    This raises an error because most-likely the dependency
    was not configured correctly.
    """

    name = "depmissingsrcs"
    expected = (
        rf"No source files found in dependency '{name}'\. "
        r"Try adjusting its config using `config_override` to get Ape to recognize the project\. "
        r"\nMore information: "
        r"https://docs.apeworx.io/ape/stable/userguides/dependencies.html#config-override"
    )
    with create_tempdir() as temp_dir:
        dependency = LocalDependency(name=name, local=str(temp_dir))

        # Raises because there are no source files in temp_dir.
        with pytest.raises(ProjectError, match=expected):
            dependency.extract_manifest()
